/*
            LTR212.
            ,    
           :
    ltr212_cbr slot crate_serial srvip
    :
        slot         -   ( 1  16)
        crate_serial -    (   )
        srvip        - ip-  ltrd  LtrServer (  )
     :    ,       

       ,    1.30.8    
       NULL    bio-     
        (Windows)     (Linux).

         .
         ,   
       .

         ,    
    .

       (  ,    
    ),   .
           Windows   CTRL+C  Linux

      VisualStudio:
          Visual Studio,     
    ( (Project) ->  (Properties) ->   (Configuration Properties)
    -> /++ ->  (General) ->    (Additional Include Directories))
            (ltr212api.h    
         ltr/include   )
        .lib   <  >/lib/msvc
    ( (Project) ->  (Properties) ->   (Configuration Properties) ->
     (Linker) ->  (General) ->    (Additional Library Directories)).

    !:      Visual Studio    
       ,    :
      (File) ->    (Advanced Save Options)...
        (Encoding)   (UTF8,  )/Unicode (UTF-8 with signature)
        .
*/


#include "ltr/include/ltr212api.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <locale.h>
#include <conio.h>
#else
#include <signal.h>
#include <unistd.h>
#include <termios.h>
#endif


#ifdef _MSC_VER
    /*     getch */
    #pragma warning( disable : 4996)
#endif


/* -   ,     */
#define RECV_CH_SIZE       4

/*      */
static int f_out = 0;

typedef struct {
    int slot;
    const char *serial;
    DWORD addr;
} t_open_param;


/*    .   ,  
 *   :
 * 1  -   ( 1  16)
 * 2  -   
 * 3  - ip-  */
static int f_get_params(int argc, char** argv, t_open_param* par) {
    int err = 0;
    par->slot = LTR_CC_CHNUM_MODULE1;
    par->serial = "";
    par->addr = LTRD_ADDR_DEFAULT;


    if (argc > 1)
        par->slot = atoi(argv[1]);
    if (argc > 2)
        par->serial = argv[2];
    if (argc > 3) {
        int a[4],i;
        if (sscanf(argv[3], "%d.%d.%d.%d", &a[0], &a[1], &a[2], &a[3])!=4) {
            fprintf(stderr, "  IP-!!\n");
            err = -1;
        }

        for (i=0; (i < 4) && !err; i++) {
            if ((a[i]<0) || (a[i] > 255)) {
                fprintf(stderr, " IP-!!\n");
                err = -1;
            }
        }

        if (!err) {
            par->addr = (a[0] << 24) | (a[1]<<16) | (a[2]<<8) | a[3];
        }
    }
    return err;
}



#ifndef _WIN32
/*     Linux */
static void f_abort_handler(int sig) {
    f_out = 1;
}


int getch(void) {
     struct termios oldt,
     newt;
     int ch;
     tcgetattr( STDIN_FILENO, &oldt );
     newt = oldt;
     newt.c_lflag &= ~( ICANON | ECHO );
     tcsetattr( STDIN_FILENO, TCSANOW, &newt );
     ch = getchar();
     tcsetattr( STDIN_FILENO, TCSANOW, &oldt );
     return ch;
}
#endif

/*   y  n   */
int get_yesno(void) {
    int ch;
    do {
        ch = getch();
    } while ((ch!='y') && (ch!='Y') && (ch!='n') && (ch!='N'));
    return (ch=='y') || (ch=='Y');
}

/*      */
static const char* f_module_type_str[] = {
    "LTR212/LTR212-M3",
    "LTR212-M1",
    "LTR212-M2"
};

/*      */
static const char* f_cbr_names[] = {
    "  ",
    "  ",
    "    ",
    "  ",
    "  ",
    "    .  ",
    "   (2-  )",
    "     . ",
};

#define CBR_MODES_CNT sizeof(f_cbr_names)/sizeof(f_cbr_names[0])

/*      */
static void f_print_ch_msk(BYTE lch_msk) {
    INT ch;
    for (ch=0; lch_msk!=0; ch++) {
        if (lch_msk & (1<<ch)) {
            printf("%d", ch+1);
            lch_msk &= ~(1<<ch);
            if (lch_msk)
                printf(", ");
        }
    }
}



int main(int argc, char** argv) {
    INT err = LTR_OK;
    TLTR212 hltr212;
    t_open_param par;
#ifndef _WIN32
    struct sigaction sa;
    /*   Linux      ,
           */
    sa.sa_handler = f_abort_handler;
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGABRT, &sa, NULL);
#endif

#ifdef _WIN32
    /*         Windows*/
    setlocale(LC_ALL, "");
#endif

    err = f_get_params(argc, argv, &par);

    if (!err) {
        LTR212_Init(&hltr212);
        /*    .    
             DSP*/
        err = LTR212_Open(&hltr212, par.addr, LTRD_PORT_DEFAULT, par.serial, par.slot, NULL);
        if (err != LTR_OK) {
            fprintf(stderr, "     .  %d (%s)\n",
                    err, LTR212_GetErrorString(err));
        } else {
            /*     ,     */
            printf("  !\n");
            printf("       = %s\n", hltr212.ModuleInfo.Name);
            printf("        = %s\n", hltr212.ModuleInfo.Serial);
            printf("            = %s\n", hltr212.ModuleInfo.Type <
                   sizeof(f_module_type_str)/sizeof(f_module_type_str[0]) ?
                        f_module_type_str[hltr212.ModuleInfo.Type] : " ");
            printf("       = %s\n", hltr212.ModuleInfo.BiosVersion);
            printf("       = %s\n",   hltr212.ModuleInfo.BiosDate);
            fflush(stdout);

            /*    EEPROM */
            err = LTR212_TestEEPROM(&hltr212);
            if (err!=LTR_OK) {
                fprintf(stderr, "    EEPROM: %d (%s)\n",
                        err, LTR212_GetErrorString(err));
            }


            if (err==LTR_OK) {
                INT ch;
                /* -------------    --------------- */
                hltr212.AcqMode = LTR212_FOUR_CHANNELS_WITH_HIGH_RESOLUTION;

                hltr212.LChQnt = 4;
                hltr212.REF = LTR212_REF_5V;
                hltr212.AC = FALSE;

                for (ch=0; ch < hltr212.LChQnt; ch++) {
                    hltr212.LChTbl[ch]=LTR212_CreateLChannel(ch+1, LTR212_SCALE_B_10);
                }
            }


            if (err==LTR_OK) {
                printf("\n    (y)    (n)?:");
                if (!get_yesno()) {
                    hltr212.UseClb = FALSE;
                    hltr212.UseFabricClb = TRUE;
                    printf("\n");
                } else {
                    int done = 0;

                    hltr212.UseClb = TRUE;
                    hltr212.UseFabricClb = FALSE;

                    while (!done) {
                        printf("\n    ? (y/n): ");
                        if (!get_yesno()) {
                            done = 1;
                            printf("\n");
                        } else {
                            int answer;
                            unsigned i;
                            printf("\n   :\n");
                            for (i=0; i < CBR_MODES_CNT; i++) {
                                printf(" %d - %s\n", i, f_cbr_names[i]);
                            }
                            printf("    :\n");
                            answer = getch();
                            if ((answer >= '0')  && (answer< (char)('0'+CBR_MODES_CNT))) {
                                INT mode = answer - '0';
                                INT reset;
                                BYTE lch_msk, lch_msk_orig;
                                INT cbr_err;
                                printf("\n   (y/n)?");
                                reset = get_yesno();

                                printf("\n   \"%s\" %s\n",
                                       f_cbr_names[mode], reset ? "  " : "  ");

                                /*   */
                                lch_msk = lch_msk_orig = (1 << hltr212.LChQnt) - 1;
                                cbr_err = LTR212_Calibrate(&hltr212, &lch_msk, mode, reset);

                                if (cbr_err == LTR212_ERR_CANT_CALIBRATE) {
                                    /*      , 
                                       lch_msk ,    */
                                    printf("    : ");
                                    f_print_ch_msk(lch_msk_orig & ~lch_msk);
                                    printf("\n");
                                    /*     , 
                                       ,     */                                    
                                    if (lch_msk) {
                                        printf("   : ");
                                        f_print_ch_msk(lch_msk);
                                        printf("\n");
                                    } else {
                                        printf("      !\n");
                                    }
                                } else if (cbr_err!=LTR_OK) {
                                    fprintf(stderr, "   !  %d (%s)\n",
                                            cbr_err, LTR212_GetErrorString(cbr_err));
                                } else {
                                    printf("  !\n");
                                }

                            } else {
                                printf("\n    ");
                            }
                        }
                    }
                }
            }




            if (err==LTR_OK) {
                err = LTR212_SetADC(&hltr212);
                if (err!=LTR_OK) {
                    fprintf(stderr, "    :  %d (%s)\n",
                        err, LTR212_GetErrorString(err));
                }
            }

            if (err==LTR_OK) {
                DWORD *wrds=NULL;
                double *data=NULL;
                INT recv_data_cnt = hltr212.LChQnt * RECV_CH_SIZE;
                //    2- 
                INT recv_wrds_cnt = 2*recv_data_cnt;
                DWORD tout;
                INT Run = 0;
                INT frame_num = 0;

                tout = LTR212_CalcTimeOut(&hltr212, RECV_CH_SIZE);

                /*        */
                wrds = (DWORD*)malloc(recv_wrds_cnt*sizeof(wrds[0]));
                data = (double*)malloc(recv_data_cnt*sizeof(data[0]));

                if ((wrds==NULL) || (data==NULL)) {
                    fprintf(stderr, "  !\n");
                    err = LTR_ERROR_MEMORY_ALLOC;
                }

                if (err==LTR_OK) {
                    err = LTR212_Start(&hltr212);
                    if (err!=LTR_OK) {
                        fprintf(stderr, "    !  %d (%s)\n",
                                err, LTR212_GetErrorString(err));
                    } else {
                        printf("  .    %s\n",
#ifdef _WIN32
                               " "
#else
                               "CTRL+C"
#endif
                               );
                        fflush(stdout);
                        Run = 1;
                    }                    
                }

                while (!f_out && (err==LTR_OK)) {
                    INT recv_cnt=0;
                    /*      */
                    recv_cnt = LTR212_Recv(&hltr212, wrds, NULL, recv_wrds_cnt, tout);
                    /*       */
                    if (recv_cnt<0) {
                        err = recv_cnt;
                        fprintf(stderr, "   :  %d (%s)\n",
                                err, LTR212_GetErrorString(err));
                    }
                    else if (recv_cnt < recv_wrds_cnt) {
                        fprintf(stderr, "  ,    !  %d,  %d\n",
                                recv_wrds_cnt, recv_cnt);
                        err = LTR_ERROR_RECV_INSUFFICIENT_DATA;
                    } else {

                        /*     */
                        err = LTR212_ProcessData(&hltr212, wrds, data, (DWORD*)&recv_cnt, TRUE);
                        if (err!=LTR_OK) {
                            printf("  !  %d (%s)\n",
                                   err, LTR212_GetErrorString(err));
                        } else {
                            int i;
                            frame_num++;
                            /*      ,
                             *      */
                            printf("   %d:  : ", frame_num);
                            for (i=0; (i < hltr212.LChQnt) && (i < recv_cnt); i++)
                                printf(" %.4f ", data[i]*1000);
                            printf("\n");
                            fflush(stdout);
                        }
                    }

#ifdef _WIN32
                    /*      */
                    if (err==LTR_OK) {
                        if (_kbhit())
                            f_out = 1;
                    }
#endif
                } // while(!f_out && (err==LTR_OK))

                /*    ,     */
                if (Run) {
                    INT stoperr = LTR212_Stop(&hltr212);
                    if (stoperr==LTR_OK) {
                        printf("  !\n");
                    } else {
                        fprintf(stderr, "   .  %d (%s)\n",
                                err, LTR212_GetErrorString(err));
                        if (err==LTR_OK)
                            err = stoperr;
                    }
                }

                free(wrds);
                free(data);
            }
        }
        LTR212_Close(&hltr212);
    }


    return err;
}
